
/*################################################################################################
 *##################                                                            ##################
 *##################                   Genetic parameters                       ##################
 *##################                                                            ##################
 *##################                     Implementation                         ##################
 *##################                                                            ##################
 *################################################################################################
 */

/*
 * RIGEL: Representation-Independant Genetic Engine Library
 *
 * Copyright (C) 2005-2005 Yann LANDRIN-SCHWEITZER a.k.a. Varkhan
 *
 *
 * This library is free software;  you can redistribute it and/or modify it
 * under  the terms of the  GNU Lesser General Public License  as published
 * by the  Free Software Foundation;  either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This  library  is  distributed  in  the hope that it will be useful, but
 * WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library;  if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Author:   Yann LANDRIN-SCHWEITZER
 * Contact:  varkhan@free.fr
 * Homepage: http://www.varkhan.net/
 */


#include <errno.h>
#include <stdlib.h>

#include <RIGEL/Parameters.h>

#include "Config.h"

#define RIGEL_PARAM_MEMINC 10



RIGEL_Parameters RIGEL_ParametersAlloc() {
	RIGEL_Parameters param ;
	param = (struct S_RIGEL_Parameters *) malloc(sizeof(struct S_RIGEL_Parameters)) ;
	if(param==NULL) {
		DbgLog("RIGEL_ParametersAlloc","RIGEL_mem",0,NULL) ;
		return NULL ;
	}
	// Default parameters: init to zero
	param->rate_init = 0.0 ;
	param->rate_mutn = 0.0 ;
	param->rate_xovr = 0.0 ;
	param->rate_migr = 0.0 ;
	param->popu_size = 0 ;	// Invalid -- should trigger reject if used
	param->popu_flag = 0 ;
	param->popu_hist = 0 ;
	param->popu_elit = 0 ;
	param->popu_parn = 0 ;
	param->popu_chld = 0 ;
	param->popu_init = 0 ;
	param->popu_fadj = 0.0 ;
	// Misc parameters hash
	param->parmem = 0 ;
	param->parnum = 0 ;
	param->parkey = NULL ;
	param->parval = NULL ;
	return param ;
}

RIGEL_Parameters RIGEL_ParametersCopy(RIGEL_Parameters param) {
	RIGEL_Parameters cpypm ;
	cpypm = (struct S_RIGEL_Parameters *) malloc(sizeof(struct S_RIGEL_Parameters)) ;
	if(cpypm==NULL) {
		DbgLog("RIGEL_ParametersCopy","RIGEL_mem",0,NULL) ;
		return NULL ;
	}
	cpypm->rate_init = param->rate_init ;
	cpypm->rate_mutn = param->rate_mutn ;
	cpypm->rate_xovr = param->rate_xovr ;
	cpypm->rate_migr = param->rate_migr ;
	cpypm->popu_size = param->popu_size ;
	cpypm->popu_flag = param->popu_flag ;
	cpypm->popu_hist = param->popu_hist ;
	cpypm->popu_elit = param->popu_elit ;
	cpypm->popu_parn = param->popu_parn ;
	cpypm->popu_chld = param->popu_chld ;
	cpypm->popu_init = param->popu_init ;
	cpypm->popu_fadj = param->popu_fadj ;
	cpypm->parmem = 0 ;
	cpypm->parnum = 0 ;
	cpypm->parkey = NULL ;
	cpypm->parval = NULL ;
	if(param->parmem>0) {
		cpypm->parkey = (char * *) malloc(param->parmem*sizeof(char *)) ;
		if(cpypm->parkey==NULL) {
			free(cpypm) ;
			DbgLog("RIGEL_ParametersCopy","RIGEL_mem",0,NULL) ;
			return NULL ;
		}
		cpypm->parval = (double *) malloc(param->parmem*sizeof(double)) ;
		if(cpypm->parval==NULL) {
			free(cpypm->parkey) ;
			free(cpypm) ;
			DbgLog("RIGEL_ParametersCopy","RIGEL_mem",0,NULL) ;
			return NULL ;
		}
	}
	cpypm->parmem = param->parmem ;
	if(param->parnum>0) {
		unsigned long i = 0 ;
		while(i<param->parnum) {
			unsigned long j = 0 ;
			char * ky ;
			while(param->parkey[i][j]!='\0') j ++ ;
			ky = (char *) malloc((j+1)*sizeof(char)) ;
			if(ky==NULL) {
				while(i>0) {
					i -- ;
					free(cpypm->parkey[i]) ;
				}
				free(cpypm->parkey) ;
				free(cpypm->parval) ;
				free(cpypm) ;
				DbgLog("RIGEL_ParametersCopy","RIGEL_mem",0,NULL) ;
				return NULL ;
			}
			ky[j] = '\0' ;
			while(j>0) {
				j -- ;
				ky[j] = param->parkey[i][j] ;
			}
			cpypm->parkey[i] = ky ;
			cpypm->parval[i] = param->parval[i] ;
			i ++ ;
		}
	}
	cpypm->parnum = param->parnum ;
	return cpypm ;
}

void RIGEL_ParametersFree(RIGEL_Parameters param) {
	if(param!=NULL) {
		unsigned long i ;
		i = param->parnum ;
		if(param->parkey!=NULL) while(i>0) {
			i -- ;
			free(param->parkey[i]) ;
			param->parkey[i] = NULL ;
		}
		free(param->parkey) ;
		param->parkey = NULL ;
		free(param->parval) ;
		param->parmem = 0 ;
		param->parnum = 0 ;
		param->popu_size = 0 ;
		free(param) ;
	}
}

double RIGEL_ParametersSetMisc(RIGEL_Parameters param, char * name, double value) {
	unsigned long i, j ;
	char * ky ;
	if(param==NULL) {
		errno = EINVAL ;
		DbgLog("RIGEL_ParametersSetMisc","RIGEL_arg",0,"NULL RIGEL_Parameters argument") ;
		return -1 ;
	}
	if(name==NULL) {
		errno = EINVAL ;
		DbgLog("RIGEL_ParametersSetMisc","RIGEL_arg",0,"NULL name string") ;
		return -1 ;
	}
	if(value<0) {
		errno = EINVAL ;
		DbgLog("RIGEL_ParametersSetMisc","RIGEL_arg",0,"Invalid negative value") ;
		return -1 ;
	}
	i = param->parnum ;
	while(i>0) {
		i -- ;
		j = 0 ;
		ky = param->parkey[i] ;
		while(name[j]!='\0' && ky[j]==name[j]) j ++ ;
		if(name[j]=='\0' && ky[j]=='\0') {
			// Found -> replace
			param->parval[i] = value ;
			return value ;
		}
	}
	// Not found -> create
	if(param->parnum>=param->parmem) {
		char * * nk ;
		double * nv ;
		nk = (char * *) realloc(param->parkey,(param->parmem+RIGEL_PARAM_MEMINC)*sizeof(char *)) ;
		if(nk==NULL) {
			DbgLog("RIGEL_ParametersSetMisc","RIGEL_mem",0,NULL) ;
			return -1 ;
		}
		param->parkey = nk ;
		nv = (double *) realloc(param->parval,(param->parmem+RIGEL_PARAM_MEMINC)*sizeof(double)) ;
		if(nv==NULL) {
			DbgLog("RIGEL_ParametersSetMisc","RIGEL_mem",0,NULL) ;
			return -1 ;
		}
		param->parval = nv ;
		param->parmem += RIGEL_PARAM_MEMINC ;
	}
	i = 0 ;
	while(name[i]!='\0') i ++ ;
	ky = (char *) malloc((i+1)*sizeof(char)) ;
	if(ky==NULL) {
		DbgLog("RIGEL_ParametersSetMisc","RIGEL_mem",0,NULL) ;
		return -1 ;
	}
	ky[i] = '\0' ;
	while(i>0) {
		i -- ;
		ky[i] = name[i] ;
	}
	param->parkey[param->parnum] = ky ;
	param->parval[param->parnum] = value ;
	param->parnum ++ ;
	return value ;
}

double RIGEL_ParametersGetMisc(RIGEL_Parameters param, char * name) {
	unsigned long i, j ;
	char * ky ;
	if(param==NULL) {
		errno = EINVAL ;
		DbgLog("RIGEL_ParametersGetMisc","RIGEL_arg",0,"NULL RIGEL_Parameters argument") ;
		return -1 ;
	}
	if(name==NULL) {
		errno = EINVAL ;
		DbgLog("RIGEL_ParametersGetMisc","RIGEL_arg",0,"NULL name string") ;
		return -1 ;
	}
	i = param->parnum ;
	while(i>0) {
		i -- ;
		j = 0 ;
		ky = param->parkey[i] ;
		while(name[j]!='\0' && ky[j]==name[j]) j ++ ;
		if(name[j]=='\0' && ky[j]=='\0') {
			// Found -> return value
			return param->parval[i] ;
		}
	}
	// Not found -> return invalid
	return -1 ;
}

RIGEL_Parameters RIGEL_ParametersSetDefaults(RIGEL_Parameters param) {
	if(param==NULL) {
		param = (struct S_RIGEL_Parameters *) malloc(sizeof(struct S_RIGEL_Parameters)) ;
		if(param==NULL) {
			DbgLog("RIGEL_ParametersSetDefaults","RIGEL_mem",0,NULL) ;
			return NULL ;
		}
		// Misc parameters hash
		param->parmem = 0 ;
		param->parnum = 0 ;
		param->parkey = NULL ;
		param->parval = NULL ;
	}
	// Default parameters
	param->rate_init = 1.0 ;
	param->rate_mutn = 1.0 ;
	param->rate_xovr = 1.0 ;
	param->rate_migr = 0.0 ;
	param->popu_size = 3 ;
	param->popu_flag = 0 ;
	param->popu_hist = 0 ;
	param->popu_elit = 1 ;
	param->popu_parn = 1 ;
	param->popu_chld = 2 ;
	param->popu_init = 0 ;
	param->popu_fadj = 0.0 ;
	return param ;
}


#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>



unsigned long RIGEL_ParametersSet(RIGEL_Parameters param, char * arg, char * val) {
	if(param==NULL) return 0 ;
	if(arg==NULL) return 0 ;
	if(val==NULL) return 0 ;
	if(!strcmp(arg,"")) {
		return 0 ;
	}
	else if(!strcmp(arg,"rate_init")) {
		double v = atof(val) ;
		if(v<0) return 0 ;
		param->rate_init = v ;
	}
	else if(!strcmp(arg,"rate_mutn")) {
		double v = atof(val) ;
		if(v<0) return 0 ;
		param->rate_mutn = v ;
	}
	else if(!strcmp(arg,"rate_xovr")) {
		double v = atof(val) ;
		if(v<0) return 0 ;
		param->rate_xovr = v ;
	}
	else if(!strcmp(arg,"rate_migr")) {
		double v = atof(val) ;
		if(v<0) return 0 ;
		param->rate_migr =  v ;
	}
	else if(!strcmp(arg,"popu_size")) {
		unsigned long v = atol(val) ;
		if(v<=0) return 0 ;
		param->popu_size = v ;
	}
	else if(!strcmp(arg,"popu_hist")) {
		unsigned long v = atol(val) ;
		if(v<0) return 0 ;
		param->popu_hist = v ;
	}
	else if(!strcmp(arg,"popu_elit")) {
		unsigned long v = atol(val) ;
		if(v<0) return 0 ;
		param->popu_elit = v ;
	}
	else if(!strcmp(arg,"popu_parn")) {
		unsigned long v = atol(val) ;
		if(v<0) return 0 ;
		param->popu_parn = v ;
	}
	else if(!strcmp(arg,"popu_chld")) {
		unsigned long v = atol(val) ;
		if(v<0) return 0 ;
		param->popu_chld = v ;
	}
	else if(!strcmp(arg,"popu_init")) {
		unsigned long v = atol(val) ;
		if(v<0) return 0 ;
		param->popu_init = v ;
	}
	else if(!strcmp(arg,"popu_flag")) {
		unsigned long i = 0 ;
		char s[16] ;
		unsigned long j ;
		param->popu_flag = 0 ;
		while(val[i]!='\0') {
			j = 0 ;
			while( (val[i+j]!='\0') && isalnum(val[i+j]) && (j<15) ) {
				s[j] = val[i+j] ;
				j ++ ;
			}
			s[j] = '\0' ;
			if(!strcmp(s,"")) {
				//Do nothing
			}
			else if(!strcmp(s,"SORTDESC")) {
				param->popu_flag |= RIGEL_PARAM_SORTDESC ;
			}
			else if(!strcmp(s,"SELPROPN")) {
				param->popu_flag |= RIGEL_PARAM_SELPROPN ;
			}
			else if(!strcmp(s,"SELRANKG")) {
				param->popu_flag |= RIGEL_PARAM_SELRANKG ;
			}
			else if(!strcmp(s,"SELTOURN")) {
				param->popu_flag |= RIGEL_PARAM_SELTOURN ;
			}
			else if(!strcmp(s,"REPPROPN")) {
				param->popu_flag |= RIGEL_PARAM_REPPROPN ;
			}
			else if(!strcmp(s,"REPRANKG")) {
				param->popu_flag |= RIGEL_PARAM_REPRANKG ;
			}
			else if(!strcmp(s,"REPTOURN")) {
				param->popu_flag |= RIGEL_PARAM_REPTOURN ;
			}
			else if(!strcmp(s,"SELCOMMA")) {
				param->popu_flag |= RIGEL_PARAM_SELCOMMA ;
			}
			else if(!strcmp(s,"SELPLUS")) {
				param->popu_flag &= ~RIGEL_PARAM_SELCOMMA ;
			}
			else if(!strcmp(s,"REPCOMMA")) {
				param->popu_flag |= RIGEL_PARAM_REPCOMMA ;
			}
			else if(!strcmp(s,"REPPLUS")) {
				param->popu_flag &= ~RIGEL_PARAM_REPCOMMA ;
			}
			i += j ;
		}
	}
	else {
		double v = atof(val) ;
		if(v<0) return 0 ;
		RIGEL_ParametersSetMisc(param,arg,v) ;
	}
	return 1 ;
}

#include <stdarg.h>

void RIGEL_ParametersParse(RIGEL_Parameters param, ...) {
	char * arg ;
	char * val ;
	va_list ap ;
	va_start(ap,param) ;
	arg = (char *) va_arg(ap, char *) ;
	while(arg!=NULL) {
		val = (char *) va_arg(ap, char *) ;
		RIGEL_ParametersSet(param,arg,val) ;
		arg = (char *) va_arg(ap, char *) ;
	}
	va_end(ap) ;
}

void RIGEL_ParametersParseArgs(RIGEL_Parameters param, int * argc, char * * * argv) {
	unsigned long i = 0 ;
	unsigned long k ;
	while((i+1)<(*argc)) {
		if((*argv)[i][0]=='+' && RIGEL_ParametersSet(param,(*argv)[i]+1,(*argv)[i+1])) {
			k = i+2 ;
			while(k<(*argc)) {
				(*argv)[k-2] = (*argv)[k] ;
				k ++ ;
			}
			(*argc) -= 2 ;
		}
		else i ++ ;
	}
}

unsigned long RIGEL_ReadParameters(RIGEL_Parameters param, int fd) {
	char eof = 0 ;
	char eol = 0 ;
	char * buf = NULL ;
	char * var ;
	char * val ;
	unsigned long mem = 0 ;
	unsigned long len = 0 ;
	unsigned long n = 0 ;
	unsigned long i, j ;
	while(!eof) {
		i = 0 ;
		while(!eol && !eof) {
			if(len>=mem) {
				buf = realloc(buf,mem+1024) ;
				if(buf==NULL) {
					DbgLog("RIGEL_ReadParameters","RIGEL_mem",0,NULL) ;
					return -1 ;
				}
				mem += 1024 ;
			}
			j = read(fd,buf+len,mem-len) ;
			if(j==-1) {
				free(buf) ;
				DbgLog("RIGEL_ReadParameters","RIGEL_mem",0,NULL) ;
				return -1 ;
			}
			if(j<(mem-len)) eof = 1 ;
			len += j ;
			while(i<len) {
				if(buf[i]=='\n') {
					eol = 1 ;
					break ;
				}
				i ++ ;
			}
		}
		// We have either an eol, or eof, and i is at the end of line / file
		buf[i] = '\0' ; i ++ ;
		j = 0 ;
		while(buf[j]==' ' && buf[j]!='\0') j ++ ;
		var = buf+j ;
		while(buf[j]!=' ' && buf[j]!='=' && buf[j]!='\0') j ++ ;
		if(buf[j]=='\0') continue ;
		buf[j] = '\0' ; j ++ ;
		while((buf[j]==' '||buf[j]=='=') && buf[j]!='\0') j ++ ;
		val = buf+j ;
		while(buf[j]!='\0') j ++ ;
		buf[j] = '\0' ;
		RIGEL_ParametersSet(param,var,val) ;
		n ++ ;
		// Clean up the buffer,moving the unused part at the beginning
		j = 0 ;
		while(i<len) {
			buf[j] = buf[i] ;
			i ++ ;
			j ++ ;
		}
		len = j ;
	}
	free(buf) ;
	return n ;
}

unsigned long RIGEL_WriteParameters(RIGEL_Parameters param, int fd) {
#define CHARNUM 512
	char s[CHARNUM] ;
	unsigned long i, n = 0 ;
	unsigned long t, r ;
	char f ;

#define PRINT_VAL(name,format) \
	t = snprintf(s,(CHARNUM-1), #name " = " format "\n",param->name) ; \
	if(t==-1) { \
		DbgLog("RIGEL_WriteParameters","RIGEL_sys",0,NULL) ; \
		return -1 ; \
	} \
	r = write(fd,s,t) ; \
	if(r==-1) { \
		DbgLog("RIGEL_WriteParameters","RIGEL_sys",0,NULL) ; \
		return -1 ; \
	} \
	n += r ; \
	if(t>=CHARNUM || r<t) { \
		return n ; \
	}

	if(param==NULL) {
		errno = EINVAL ;
		DbgLog("RIGEL_WriteParameters","RIGEL_arg",0,"NULL parameters argument") ;
		return -1 ;
	}
	if(fd==-1) {
		errno = EINVAL ;
		DbgLog("RIGEL_WriteParameters","RIGEL_arg",0,"Invalid file descriptor") ;
		return -1 ;
	}
	PRINT_VAL(popu_size,"%ld") ;
	PRINT_VAL(popu_hist,"%ld") ;
	PRINT_VAL(popu_elit,"%ld") ;
	PRINT_VAL(popu_parn,"%ld") ;
	PRINT_VAL(popu_chld,"%ld") ;
	PRINT_VAL(popu_init,"%ld") ;

	f = ' ' ;
	t = 0 ;
	strncpy(s+t,"popu_flag = ",(CHARNUM-1)) ;
	t += 12 ;
	if(t<(CHARNUM-1) && param->popu_flag&RIGEL_PARAM_SORTDESC) {
		s[t] = f ; t ++ ;
		strncpy(s+t,"SORTDESC",(CHARNUM-1-t)) ;
		t += 8 ;
		f = '|' ;
	}
	if(t<(CHARNUM-1) && param->popu_flag&RIGEL_PARAM_SELPROPN) {
		s[t] = f ; t ++ ;
		strncpy(s+t,"SELPROPN",(CHARNUM-1-t)) ;
		t += 8 ;
		f = '|' ;
	}
	if(t<(CHARNUM-1) && param->popu_flag&RIGEL_PARAM_SELRANKG) {
		s[t] = f ; t ++ ;
		strncpy(s+t,"SELRANKG",(CHARNUM-1-t)) ;
		t += 8 ;
		f = '|' ;
	}
	if(t<(CHARNUM-1) && param->popu_flag&RIGEL_PARAM_SELTOURN) {
		s[t] = f ; t ++ ;
		strncpy(s+t,"SELTOURN",(CHARNUM-1-t)) ;
		t += 8 ;
		f = '|' ;
	}
	if(t<(CHARNUM-1) && param->popu_flag&RIGEL_PARAM_REPPROPN) {
		s[t] = f ; t ++ ;
		strncpy(s+t,"REPPROPN",(CHARNUM-1-t)) ;
		t += 8 ;
		f = '|' ;
	}
	if(t<(CHARNUM-1) && param->popu_flag&RIGEL_PARAM_REPRANKG) {
		s[t] = f ; t ++ ;
		strncpy(s+t,"REPRANKG",(CHARNUM-1-t)) ;
		t += 8 ;
		f = '|' ;
	}
	if(t<(CHARNUM-1) && param->popu_flag&RIGEL_PARAM_REPTOURN) {
		s[t] = f ; t ++ ;
		strncpy(s+t,"REPTOURN",(CHARNUM-1-t)) ;
		t += 8 ;
		f = '|' ;
	}
	if(t<(CHARNUM-1)) { if(param->popu_flag&RIGEL_PARAM_SELCOMMA) {
		s[t] = f ; t ++ ;
		strncpy(s+t,"SELCOMMA",(CHARNUM-1-t)) ;
		t += 8 ;
		f = '|' ;
	}
	else {
		s[t] = f ; t ++ ;
		strncpy(s+t,"SELPLUS",(CHARNUM-1-t)) ;
		t += 7 ;
		f = '|' ;
	} }
	if(t<(CHARNUM-1)) { if(param->popu_flag&RIGEL_PARAM_REPCOMMA) {
		s[t] = f ; t ++ ;
		strncpy(s+t,"REPCOMMA",(CHARNUM-1-t)) ;
		t += 8 ;
		f = '|' ;
	}
	else {
		s[t] = f ; t ++ ;
		strncpy(s+t,"REPPLUS",(CHARNUM-1-t)) ;
		t += 7 ;
		f = '|' ;
	} }
	s[t] = '\n' ; t ++ ;
	r = write(fd,s,t) ;
	if(r==-1) {
		DbgLog("RIGEL_WriteParameters","RIGEL_sys",0,NULL) ;
		return -1 ;
	}
	n += r ;
	if(t>=CHARNUM || r<t) {
		return n ;
	}

	// Printing rates
	PRINT_VAL(rate_init,"%f") ;
	PRINT_VAL(rate_mutn,"%f") ;
	PRINT_VAL(rate_xovr,"%f") ;
	PRINT_VAL(rate_migr,"%f") ;

	// Printing misc
	i = 0 ;
	while(i<param->parnum) {
		t = snprintf(s,(CHARNUM-1),"%s = %f\n",param->parkey[i],param->parval[i]) ;
		if(t==-1) {
			DbgLog("RIGEL_WriteParameters","RIGEL_sys",0,NULL) ;
			return -1 ;
		}
		r = write(fd,s,t) ;
		if(r==-1) {
			DbgLog("RIGEL_WriteParameters","RIGEL_sys",0,NULL) ;
			return -1 ;
		}
		n += r ;
		if(t>=CHARNUM || r<t) {
			return n ;
		}
		i ++ ;
	}
	return n ;
#undef CHARNUM
#undef PRINT_VAL
}



