#!/usr/bin/perl

################################################################################
###                                                                          ###
###                  TextDoc Format Parser and Transducer                    ###
###                                                                          ###
###                             (C) 2005 Varkhan                             ###
###                                                                          ###
################################################################################


# TextDoc.pm: TextDoc Format Parser and Transducer methods
#
# Copyright (C) 2001-2004 Yann LANDRIN-SCHWEITZER, a.k.a. Varkhan
#
#
# This library is free software;  you can redistribute it and/or modify it
# under  the terms of the  GNU Lesser General Public License  as published
# by the  Free Software Foundation;  either version 2.1 of the License, or
# (at your option) any later version.
#
# This  library  is  distributed  in  the hope that it will be useful, but
# WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library;  if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#
# Author:   Yann LANDRIN-SCHWEITZER, a.k.a. Varkhan
# Contact:  varkhan@varkhan.net
# Homepage: http://www.varkhan.net/



package TextDoc ;

BEGIN {
	use Exporter	();
	use vars	qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

	$VERSION	= 0.1 ;						       
#       $VERSION	= do{my@r=q$Revision: 0.1 $=~/\d+/g;sprintf '%d.$trans%02d'x$#r,@r};

	@ISA		= qw(Exporter);
	@EXPORT		= qw(&TransFile &SetTrans &SetHeader &SetTailer);
	@EXPORT_OK	= qw(&TransFile &SetTrans &SetHeader &SetTailer &NextBlock &Protoize);
	%EXPORT_TAGS	= (
		'FILES'		=> qw(&TransFile &SetTrans &SetHeader &SetTailer),
		'PARSE'		=> qw(&NextBlock),
		'PROTO'		=> qw(&Protoize)
			 ) ;
}									     



################################################################################################
#####   
#####   Prototype parser
#####



sub Parse_Type($) {
	my $proto = shift ;
	my $rght = $proto ;
	my @type = () ;
	while(length($rght)) {
		my $left ;
#		print "### ",join("/",@type)," \"$rght\"\n" ;
		if($rght=~/^\s+/) { (undef,$rght) = split(/^\s+/,$rght,2) ; }
		# Storage modifier
		if($rght=~/^(?:(?:static)|(?:extern)|(?:const)|(?:volatile)|(?:register))[\s\(\)\{,]/) {
			($left,$rght) = split(/(?=[\s\(\{,])/,$rght,2) ;
#			print "**1 \"$left\" - \"$rght\"\n" ;
		}
		# Inline structure definition
		elsif($rght=~/^struct[\s\(\{,]/) {
			($left,$rght) = split(/^(struct\s+\w+)(?=[\s\(\)\{,])/,$rght,2) ;
#			print "**2 \"$left\" - \"$rght\"\n" ;
		}
		# Function type
		elsif($rght=~/^\(\s*\*/) {
			($t,$left,$rght) = split(/^(\(\s*\*\s*\w+\)\s*\([^\)]*\))/,$rght,3) ;
#			print "**3 \"$left\" - \"$rght\"\n" ;
		}
		# Pointer
		elsif($rght=~/^\*[\s\(\)\{,]/) {
			($t,$left,$rght) = split(/^(\*)(?=[\s\(\)\{,])/,$rght,3) ;
#			print "**4 \"$left\" - \"$rght\"\n" ;
		}
		# Bare word (maybe name?)
		elsif($rght=~/^\w+[\s\(\)\{,]/) {
			($t,$left,$rght) = split(/^(\w+)(?=[\s\(\)\{,])/,$rght,3) ;
#			print "**5 \"$left\" - \"$rght\"\n" ;
		}
		else {
#			print "!!! \"$rght\"\n" ;
			last ;
		}
		push @type, $left ;
	}
	if($rght!~/^\s*[\(\)\{,;]/) {
		return (undef,undef,$rght) ;
	}
	my $name = pop @type ;
	if($name!~/^\w+$/) {
		return (undef,undef,$rght) ;
	}
	return (\@type,$name,$rght) ;
}


sub Protoize($) {
	$_ = shift ;

	# Parsing function class
	my @clas = () ;
	while(length($_)) {
		if(/^(?:inline)\s/) {
			my $left ;
			($left,$_) = split(/\s/,$_,2) ;
			push @clas, $left ;
		}
		else { last ; }
	}
	my $clas = join(' ',@clas) ;
	# Parsing function type
	my($type,$name) ;
	($type,$name,$_) = Parse_Type($_) ;
	if(!defined($type) || !defined($name)) { return undef ; }
	my $args = undef ;
	if(/^\s*\(/) {
		$args = [] ;
		# Function prototype
		(undef,$_) = split(/\(/,$_,2) ;
		while(length($_) && !/^\s*\)/) {
			my ($argt,$argn) ;
			($argt,$argn,$_) = Parse_Type($_) ;
			# Last parenthesis => finished
			if(/^\s*\)/) {
			push @{$args}, [ $argt, $argn ] ;
				last ;
			}
			# Coma => clear and next arg
			elsif(/^\s*,/) {
				(undef,$_) = split(/^\s*,/,$_,2) ;
			}
			# Var arg
			elsif(/\.\.\.\s*\)/) {
				push @{$args}, [ undef, '...' ] ;
				last ;
			}
			else { return undef ; }
			push @{$args}, [ $argt, $argn ] ;
		}
		(undef,$_) = split(/\)/,$_,2) ;
	}
	if(!/^\s*;/) { return undef ; }
	return ( $clas, $type, $name, $args ) ;
}



################################################################################################
#####
#####   Block parser
#####



my $BlockTypes = {
	'head'	=> [ '##################' , \&ParseBlock_Head ] ,
	'sect'	=> [ '#####',		\&ParseBlock_Sect ] ,
	'comm'	=> [ '#',		\&ParseBlock_Comm ] ,
	'func'	=> [ 'Function',	\&ParseBlock_Func ] ,
	'macr'	=> [ 'Macro',		\&ParseBlock_Macr ] ,
	'type'	=> [ 'Typedef',		\&ParseBlock_Func ] ,
	'note'	=> [ 'Note',		\&Parse_Note ] ,
} ;


sub NextBlock($) {
	my $fh = shift ;
	# Clear empty lines
	while(<$fh>) {
		chomp ;
		if(!/^\s*$/) { last ; }
	}
	foreach my $type (keys %{$BlockTypes}) {
		my ($match,$handl) = @{$BlockTypes->{$type}} ;
		if(/^$match\s/) { return $handl->($fh,$type,$_) ; }
	}
	return {} ;
}


sub Parse_Enum($$) {
	my $file = shift ;
	$_ = shift ;

	if(!/^  (\w+):\s*$/) { die "ParseBlock_Enum: did not find enum name" ; }
	my $name = $1 ;

	my @enum = ( ) ;
	my %enum = ( ) ;
	my $item = undef ;
	my $cont ;
	while(<$file>) {
		chomp ;
		# Empty line => end of enum
		if(/^\s*$/) { last ; }
		# Item line
		elsif(/^(?: {3,})|(?:\s*\t\s*)_/) {
			(undef,$_) = split(/^\s+_\s*/,$_,2) ;
			($item,$cont) = split(/\s*:\s*/,$_,2) ;
			push(@enum,$item) ;
			$enum{$item} = $cont ;
		}
		# Continuation line
		elsif(/^(?: {3,})|(?:\s*\t\s*)/) {
			if(!defined($item)) { die "ParseBlock_Enum: continuation line before first item" ; }
			(undef,$_) = split(/^\s+/,$_,2) ;
			$enum{$item} .= $_ ;
		}
		# Invalid line (margin regression)
		else { die "ParseBlock_Enum: non-enum lines before end of enum" ; }
	}
	return [ $name, \%enum ] ;
}


sub Parse_Note($$) {
	my $file = shift ;
	my $line = shift ;

	if(!/^  (\w+):\s*$/) { die "Parse_Note: did not find note name" ; }
	my $name = $1 ;

	my @note = () ;
	while(<$file>) {
		chomp ;
		# Empty line => end of note
		if(/^\s*$/) { last ; }
		# Item line
		elsif(/^(?:(?: {3,})|(?:\s*\t\s*))_/) {
			(undef,$line) = split(/^\s+_\s*/,$_,2) ;
			my @ent = split(/\s*:\s*/,$line) ;
			push(@note,\@ent) ;
		}
		# Note line
		elsif(/^(?: {3,})|(?:\s*\t\s*)/) {
			(undef,$line) = split(/^\s+/,$_,2) ;
			push(@note,$line) ;
		}
		# Unaligned line => error
		else { die "Parse_Note: non-aligned line before end of note" ; }
	}
	return [ $name, \@note ] ;
}


sub ParseBlock_Func($$$) {
	my $file = shift ;
	my $type = shift ;
	my $line = shift ;

	my (undef,$name) = split(/\s+/,$line,2) ;
	$line = <$file> ; chomp $line ;
	if($line!~/^=+$/) { die "ParseBlock_Func: did not find head delimiter" ; }
	$line = <$file> ; chomp $line ;
	if($line!~/^\s*$/) { die "ParseBlock_Func: did not find head spacer" ; }
	$line = <$file> ; chomp $line ;
	if($line!~/;\s*$/) { die "ParseBlock_Func: did not find valid prototype" ; }
	my $prot = $line ;
	$line = <$file> ; chomp $line ;
	if($line!~/.\s*$/) { die "ParseBlock_Func: did not find valid description" ; }
	my $desc = $line ;
	$line = <$file> ; chomp $line ;
	if($line!~/^\s*$/) { die "ParseBlock_Func: did not find description spacer" ; }
	my $blanks = 1 ;
	my ($args,$rets,$errs,$note) ;
	while(<$file>) {
		if(/^\s*$/) {
			$blanks ++ ;
			if($blanks>2) { last ; }
		}
		elsif(/^  Args:\s*$/) {
			$args = Parse_Note($file,$_) ;
			$blanks = 1 ;
		}
		elsif(/^  Returns:\s*$/) {
			$rets = Parse_Note($file,$_) ;
			$blanks = 1 ;
		}
		elsif(/^  Errors:\s*$/) {
			$errs = Parse_Note($file,$_) ;
			$blanks = 1 ;
		}
		elsif(/^  Note:\s*$/) {
			$note = Parse_Note($file,$_) ;
			$blanks = 1 ;
		}
		elsif(/^  \w+:\s*$/) {
			$note = Parse_Note($file,$_) ;
			$blanks = 1 ;
		}
		else {
			$note = [ 'OOPS', [ "Something was wrong here" ] ] ;
		}
	}
	return {
		'type'	=> $type ,
		'name'	=> $name ,
		'prot'	=> $prot ,
		'desc'	=> $desc ,
		'args'	=> $args ,
		'rets'	=> $rets ,
		'errs'	=> $errs ,
		'note'	=> $note
	} ;
}

sub ParseBlock_Macr($$$) {
	my $file = shift ;
	my $type = shift ;
	my $line = shift ;

	my (undef,$name) = split(/\s+/,$line,2) ;
	$_ = <$file> ; chomp ;
	if(!/^=+$/) { die "ParseBlock_Macr: did not find head delimiter" ; }
	$_ = <$file> ; chomp ;
	if(!/^\s*$/) { die "ParseBlock_Macr: did not find head spacer" ; }
	$_ = <$file> ; chomp ;
#	if(!/;\s*$/) { die "ParseBlock_Macr: did not find valid prototype" ; }
	my $prot = $_ ;
	$_ = <$file> ; chomp ;
	if(!/.\s*$/) { die "ParseBlock_Macr: did not find valid description" ; }
	my $desc = $_ ;
	$_ = <$file> ; chomp ;
	if(!/^\s*$/) { die "ParseBlock_Macr: did not find description spacer" ; }
	my $blanks = 1 ;
	my ($args,$rets,$note) ;
	while(<$file>) {
		if(/^\s*$/) {
			$blanks ++ ;
			if($blanks>2) { last ; }
		}
		elsif(/^  Args:\s*$/) {
			$args = Parse_Note($file,$_) ;
			$blanks = 1 ;
		}
		elsif(/^  Returns:\s*$/) {
			$rets = Parse_Note($file,$_) ;
			$blanks = 1 ;
		}
		elsif(/^  Note:\s*$/) {
			$note = Parse_Note($file,$_) ;
			$blanks = 1 ;
		}
	}
	return {
		'type'	=> $type ,
		'name'	=> $name ,
		'prot'	=> $prot ,
		'desc'	=> $desc ,
		'args'	=> $args ,
		'rets'	=> $rets ,
		'errs'	=> undef ,
		'note'	=> $note
	} ;
}

sub ParseBlock_Type($$$) {
	my $file = shift ;
	my $type = shift ;
	my $line = shift ;

	my (undef,$name) = split(/\s+/,$line,2) ;
	$_ = <$file> ; chomp ;
	if(!/^=+$/) { die "ParseBlock_Type: did not find head delimiter" ; }
	$_ = <$file> ; chomp ;
	if(!/^\s*$/) { die "ParseBlock_Type: did not find head spacer" ; }
	$_ = <$file> ; chomp ;
	if(!/;\s*$/) { die "ParseBlock_Type: did not find valid prototype" ; }
	my $prot = $_ ;
	$_ = <$file> ; chomp ;
	if(!/.\s*$/) { die "ParseBlock_Type: did not find valid description" ; }
	my $desc = $_ ;
	$_ = <$file> ; chomp ;
	if(!/^\s*$/) { die "ParseBlock_Type: did not find description spacer" ; }
	my $blanks = 1 ;
	my $note ;
	while(<$file>) {
		if(/^\s*$/) {
			$blanks ++ ;
			if($blanks>2) { last ; }
		}
		elsif(/^  Note:\s*$/) {
			$note = Parse_Note($file,$_) ;
			$blanks = 1 ;
		}
	}
	return {
		'type'	=> $type ,
		'name'	=> $name ,
		'prot'	=> $prot ,
		'desc'	=> $desc ,
		'args'	=> undef ,
		'rets'	=> undef ,
		'errs'	=> undef ,
		'note'	=> $note
	} ;
}


sub ParseBlock_Head($$$) {
	my $file = shift ;
	my $type = shift ;
	my $line = shift ;

	$_ = $line ;
	my $note = '' ;
	if(!/#+$/) {
		(undef,$_) = split(/^#*\s*/,$_,2) ;
		($_,undef) = split(/\s*#*$/,$_,2) ;
		$note = $_ ;
	}
	while(<$file>) {
		chomp ;
		if(/^\s*$/) { last ; }
		elsif(!/^#/) { last ; }
		elsif(/^#+$/) { next ; }
		else {
			(undef,$_) = split(/^#*\s*/,$_,2) ;
			($_,undef) = split(/\s*#*$/,$_,2) ;
		}
		$note .= $_."\n" ;
	}
	return {
		'type'	=> $type ,
		'name'	=> $note ,
		'prot'	=> undef ,
		'desc'	=> $note ,
		'args'	=> undef ,
		'rets'	=> undef ,
		'errs'	=> undef ,
		'note'	=> [ 'Header', $note ]
	} ;
}


sub ParseBlock_Sect($$$) {
	my $file = shift ;
	my $type = shift ;
	my $line = shift ;

	$_ = $line ;
	my $note = '' ;
	if(!/#+$/) {
		(undef,$_) = split(/^#*\s*/,$_,2) ;
		($_,undef) = split(/\s*#*$/,$_,2) ;
		$note = $_ ;
	}
	while(<$file>) {
		chomp ;
		if(/^\s*$/) { last ; }
		elsif(!/^#/) { last ; }
		elsif(/^#+$/) { next ; }
		else {
			(undef,$_) = split(/^#*\s*/,$_,2) ;
			($_,undef) = split(/\s*#*$/,$_,2) ;
		}
		$note .= $_."\n" ;
	}
	return {
		'type'	=> $type ,
		'name'	=> $note ,
		'prot'	=> undef ,
		'desc'	=> $note ,
		'args'	=> undef ,
		'rets'	=> undef ,
		'errs'	=> undef ,
		'note'	=> [ 'Section', $note ]
	} ;
}


sub ParseBlock_Comm($$$) {
	my $file = shift ;
	my $type = shift ;
	my $line = shift ;

	$_ = $line ;
	my $note = '' ;
	if(!/#+$/) {
		(undef,$_) = split(/^#*\s*/,$_,2) ;
		($_,undef) = split(/\s*#*$/,$_,2) ;
		$note = $_ ;
	}
	while(<$file>) {
		chomp ;
		if(/^\s*$/) { last ; }
		elsif(!/^#/) { last ; }
		else {
			(undef,$_) = split(/^#*\s*/,$_,2) ;
		}
		$note .= $_."\n" ;
	}
	return {
		'type'	=> $type ,
		'name'	=> $note ,
		'prot'	=> undef ,
		'desc'	=> $note ,
		'args'	=> undef ,
		'rets'	=> undef ,
		'errs'	=> undef ,
		'note'	=> [ 'Coment', $note ]
	} ;
}


sub ParseBlock_Note($$$) {
	my $file = shift ;
	my $type = shift ;
	my $line = shift ;

	my (undef,$name) = split(/\s+/,$line,2) ;
	$_ = <$file> ; chomp ;
	if(!/^\s*$/) { die "ParseBlock_Note: did not find head spacer" ; }
	my $note = '' ;
	while(<$file>) {
		chomp ;
		if(/^\s*$/) { last ; }
		elsif(/^(?: {3,})|(?:\s*\t\s*)/) {
			$note .= $_ ;
		}
		last ;
	}
	return {
		'type'	=> $type ,
		'name'	=> $name ,
		'prot'	=> undef ,
		'desc'	=> undef ,
		'args'	=> undef ,
		'rets'	=> undef ,
		'errs'	=> undef ,
		'note'	=> [ 'Note', $note ]
	} ;
}


sub ShowBlock($) {
	my $block = shift ;
	foreach my $key (keys %{$block}) {
		if(ref($block->{$key})eq'HASH') {
			print "$key => {\n" ;
			foreach my $key2 (keys %{$block->{$key}}) {
				print "\t$key2 => ",$block->{$key}->{$key2},"\n" ;
			}
			print "\t}\n" ;
		}
		elsif(ref($block->{$key})eq'ARRAY') {
			print "$key => [\n" ;
			foreach my $item (@{$block->{$key}}) {
				print "\t",$item,"\n" ;
			}
			print "\t]\n" ;
		}
		elsif(defined($block->{$key})) {
			print "$key => ",$block->{$key},"\n" ;
		}
		else {
			print "$key => undef\n" ;
		}
	}
}



################################################################################################
#####
#####   Global parser
#####


sub Print_Note($$) {
	my $name = shift ;
	my $note = shift ;
	print "  ",$name,":\n" ;
	foreach my $line (@{$note}) {
		if(ref($line)eq'ARRAY') {
			my @line = @{$line} ;
			print "\t_",shift(@line) ;
			foreach my $item (@line) {
				print "\t: ",$item ;
			}
			print "\n" ;
		}
		else {
			print "\t",$line,"\n" ;
		}
	}
}


sub PrintBlock_Func($) {
	my $block = shift ;
	print "Function ",$block->{'name'},"\n" ;
	print '=' x (length($block->{'name'})+9), "\n" ;
	print "\n" ;
	print $block->{'prot'},"\n" ;
	if($block->{'desc'}=~/\.$/) {
		print $block->{'desc'},"\n" ;
	}
	else {
		print $block->{'desc'},".\n" ;
	}
	print "\n" ;
	if(defined $block->{'args'}) {
		Print_Note($block->{'args'}->[0],$block->{'args'}->[1]) ;
		print "\n" ;
	}
	if(defined $block->{'rets'}) {
		Print_Note($block->{'rets'}->[0],$block->{'rets'}->[1]) ;
		print "\n" ;
	}
	if(defined $block->{'errs'}) {
		Print_Note($block->{'errs'}->[0],$block->{'errs'}->[1]) ;
		print "\n" ;
	}
	if(defined $block->{'note'}) {
		Print_Note($block->{'note'}->[0],$block->{'note'}->[1]) ;
		print "\n" ;
	}
	print "\n" ;
}


sub PrintBlock_Macr($) {
	my $block = shift ;
	print "Macro ",$block->{'name'},"\n" ;
	print '=' x (length($block->{'name'})+6), "\n" ;
	print "\n" ;
	print $block->{'prot'},"\n" ;
	if($block->{'desc'}=~/\.$/) {
		print $block->{'desc'},"\n" ;
	}
	else {
		print $block->{'desc'},".\n" ;
	}
	print "\n" ;
	if(defined $block->{'args'}) {
		Print_Note($block->{'args'}->[0],$block->{'args'}->[1]) ;
		print "\n" ;
	}
	if(defined $block->{'rets'}) {
		Print_Note($block->{'rets'}->[0],$block->{'rets'}->[1]) ;
		print "\n" ;
	}
	if(defined $block->{'note'}) {
		Print_Note($block->{'note'}->[0],$block->{'note'}->[1]) ;
		print "\n" ;
	}
	print "\n" ;
}


sub PrintBlock_Type($) {
	my $block = shift ;
	print "Typedef ",$block->{'name'},"\n" ;
	print '=' x (length($block->{'name'})+8), "\n" ;
	print "\n" ;
	print $block->{'prot'},"\n" ;
	if($block->{'desc'}=~/\.$/) {
		print $block->{'desc'},"\n" ;
	}
	else {
		print $block->{'desc'},".\n" ;
	}
	print "\n" ;
	Print_Note($block->{'note'}->[0],$block->{'note'}->[1]) ;
	print "\n" ;
	print "\n" ;
}

sub PrintBlock_Head($) {
	my $block = shift ;
	print "#" x 80, "\n" ;
	print "#" x 18, " " x 44, "#" x 18, "\n" ;
	my @lines = split(/\n/,$block->{'note'}[1]) ;
	foreach my $line (@lines) {
		my $len1 = int((44-length($line))/2) ;
		my $len2 = 44-length($line)-$len1 ;
		print "#" x 18, " " x $len1, $line, " " x $len2, "#" x 18, "\n" ;
	}
	print "#" x 18, " " x 44, "#" x 18, "\n" ;
	print "#" x 80, "\n" ;
	print "\n" ;
}

sub PrintBlock_Sect($) {
	my $block = shift ;
	print "#" x 80, "\n" ;
	print "#" x 5, "\n" ;
	my @lines = split(/\n/,$block->{'note'}->[1]) ;
	foreach my $line (@lines) {
		print "#" x 5, " " x 3, $line, "\n" ;
	}
	print "#" x 5, "\n" ;
	print "\n" ;
}

sub PrintBlock_Comm($) {
	my $block = shift ;
	my @lines = split(/\n/,$block->{'note'}[1]) ;
	foreach my $line (@lines) {
		print "# ", $line, "\n" ;
	}
	print "\n" ;
}

sub PrintBlock_Note($) {
	my $block = shift ;
#	print "Note ",$block->{'name'},"\n" ;
	print "Note\n" ;
	print '=' x (length($block->{'name'})+5), "\n" ;
	print "\n" ;
	my @lines = split(/\n/,$block->{'note'}[1]) ;
	foreach my $line (@lines) {
		print "\t", $line, "\n" ;
	}
	print "\n" ;
}


my $TransForm = {
	'head'	=> \&PrintBlock_Head ,
	'sect'	=> \&PrintBlock_Sect ,
	'comm'	=> \&PrintBlock_Comm ,
	'func'	=> \&PrintBlock_Func ,
	'macr'	=> \&PrintBlock_Macr ,
	'type'	=> \&PrintBlock_Type ,
	'note'	=> \&PrintBlock_Note ,
} ;


my $TransHead = '' ;

my $TransTail = '' ;


sub SetTrans($$) {
	my $type = shift ;
	my $hdlr = shift ;
	if(exists $BlockTypes->{$type}) {
		$TransForm->{$type} = $hdlr ;
	}
}

sub SetHeader($) {
	$TransHead = shift ;
}

sub SetTailer($) {
	$TransTail = shift ;
}


sub TransFile($) {
	my $file = shift ;
	my $fh ;
	open($fh,'<',$file) ;
	print $TransHead, "\n" ;
	while(!eof($fh)) {
		my $block = NextBlock($fh) ;
		if(exists $block->{'type'}) {
			$TransForm->{$block->{'type'}}($block) ;
		}
		print "\n" ;
	}
	print $TransTail, "\n" ;
}



### END
1 ;
